
    ::selection {
      color: var(--dark-6);
      background-color: var(--yellow-4);
    }

    @media (prefers-color-scheme: dark) {
      ::selection {
        background-color: var(--yellow-1);
      }
    }
