# SPDX-FileCopyrightText: 2022 Chris Heywood
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk


@Gtk.Template(resource_path="/org/gnome/World/Iotas/ui/index_menu_button.ui")
class IndexMenuButton(Gtk.MenuButton):
    __gtype_name__ = "IndexMenuButton"
