    :root {
      --text-colour: %s;
      --figcaption-opacity: %s;
      --ticked-checkbox: %s;
    }
    @media screen {
      html {
        font-size: %dpt;
      }
      body {
        max-width: %dpx;
        font-family: %s;
        padding-top: %dpx;
        padding-bottom: %dpx;
      }
    }
    @media screen and (max-width:600px) {
      body {
        padding-top: %s;
      }
    }
    @media (prefers-color-scheme: dark) {
      :root {
        --text-colour: %s;
        --ticked-checkbox: %s;
      }
    }
